@@ -1,77 +1,64 @@
 # Built in Modules
-from importlib import resources as impresources
+import importlib
 import itertools
 import sys
 # Local package imports
-from . import data
+
+# from . import profilers
 
 
 class Script:
     """
-        Bash class to read existing bash scripts, pull out the options and create new file to be submitted so the
-        work in the script gets profiled.
+    Class to read existing bash scripts, pull out the options and create an object that contains all the
+    queue options, bash options, and desired profiling methods to be injected
 
-        Parameters to initiate
-        ----------
-        queue_system : str
-            str of the queue system to be used. [sbatch, torque]
-        work_script : str
-            str of the path and name of the original bash script to be profiled
-        read_queue_system : str = None
-            str of the queue system that the work_script was written for, if not specified it is assumed
-            it is the same as queue_system. [sbatch, torque]
-        queue_options : dict = None
-            dictionary of options to be passed to the queue system.
-        likwid : bool = False
-            bool determining if likwid should be used to measure the performance such that a roofline
-            model can be plotted
-        likwid_req: list = None
-            list of the required lines that need to be added to a bash script in order to allow likwid
-            to be used. For example, lines that load the likwid module if modules is being used on the
-            system.
-        prometheus : bool = False
-            bool determining if prometheus should be used to measure the metrics of the node on which
-            the script is being run.
-        prometheus_req: list = None
-            list of the required lines that need to be added to a bash script in order to allow prometheus
-            to be used. In order to run prometheus this list will need at least one entry:
-                'export PROMETHEUS_SOFTWARE=<path to the Prometheus software to be used>'
-            with an optional entry to change the python executable to use by declaring:
-                'export PROMETHEUS_PYTHON=<path to the Python version to be used to scrape the prometheus database>'
-            any additional entries will be added, but are at users discretion
+    Parameters to initiate
+    ----------
+    queue_system : str
+        str of the queue system to be used. [sbatch, torque]
+    work_script : str
+        str of the path and name of the original bash script to be profiled
+    read_queue_system : str = None
+        str of the queue system that the work_script was written for, if not specified it is assumed
+        it is the same as queue_system. [sbatch, torque]
+    queue_options : dict = None
+        dictionary of options to be passed to the queue system.
+    profiling : dict = None
+        dictionary where keys are the name of the profiler to use, and the values are dictionaries containing
+        "requirements" or other optional commands depending on the profiler being used.
 
-        Notes
-        -----
-        The Script class is intended to manage the components necessary in order to take an existing bash script,
-        extract any options for a queue system it has, add any desired options, and initiate the information needed
-        for prometheus and likwid to perform their metric collections. This object can then be passed on to the
-        submit() function in order to create the profiling bash script and submit it to the queue that it is
-        designed to be submitted to.
+    Notes
+    -----
+    The Script class is intended to manage the components necessary in order to take an existing bash script,
+    extract any options for a queue system it has, add any desired options, and initiate the information needed
+    for prometheus and likwid to perform their metric collections. This object can then be passed on to the
+    submit() function in order to create the profiling bash script and submit it to the queue that it is
+    designed to be submitted to.
 
-        Adding new queue systems
-        -----
-        To add a new queue system, go to the sections marked Queue System specifics {1}, {2}, {3} and {4} in order to
-        follow the instructions there
+    Adding new queue systems
+    -----
+    To add a new queue system, go to the sections marked Queue System specifics {1}, {2}, {3} and {4} in order to
+    follow the instructions there
 
-        Examples
-        -------- # As of now this example doesn't apply to this class.
-        #>>> from PyProfQueue.script import Script
-        #>>> ProfileScript = Script(queue_system='slurm',
-                                    work_script='./example.sh',
-                                    read_queue_system='slurm',
-                                    likwid=True,
-                                    likwid_req=['module load oneAPI_comp',
-                                                'module load likwid'],
-                                    prometheus=True,
-                                    prometheus_req=['export PROMETHEUS_SOFTWARE=~/Software/prometheus'],
-                                    queue_options={'user': 'user_name'})
+    Examples
+    -------- # As of now this example doesn't apply to this class.
+    #>>> from PyProfQueue.script import Script
+    #>>> ProfileScript = Script(queue_system='slurm',
+                                work_script='./example.sh',
+                                read_queue_system='slurm',
+                                likwid=True,
+                                likwid_req=['module load oneAPI_comp',
+                                            'module load likwid'],
+                                prometheus=True,
+                                prometheus_req=['export PROMETHEUS_SOFTWARE=~/Software/prometheus'],
+                                queue_options={'user': 'user_name'})
     """
     def __init__(self, queue_system: str,
                  work_script: str,
                  read_queue_system: str = None,
                  queue_options: dict = None,
-                 likwid: bool = False, likwid_req: list = None,
-                 prometheus: bool = False, prometheus_ip: str = None, prometheus_req: list = None):
+                 profiling: dict = None
+                 ):
         if True:  # If statement added to allow for the collapse of the initiation of variables
             self.queue_system = queue_system
             self.work_script = work_script
@@ -79,21 +66,10 @@ class Script:
             self.tmp_profile_script = None
             self.works = None
             self.work_dir = None
-            self.likwid = likwid
-            self.likwid_file = None
-            self.likwid_initEndSplit = None
-            self.likwid_req = None
-            self.prometheus = prometheus
-            self.prometheus_ip = prometheus_ip
-            self.prometheus_file = None
-            self.prometheus_initEndSplit = None
+            self.profiling = profiling
+            self.at_execute = False  # boolean to see if a profiler is already used at the execution line.
             self.read_queue_system = read_queue_system
 
-        if self.likwid:
-            self.add_likwid(likwid_req)
-        if self.prometheus:
-            self.add_prometheus(prometheus_req)
-
         if bool(queue_options):
             self.obj_options = self.Options(queue_options)
         else:
@@ -114,7 +90,7 @@ class Script:
         - Second, add a case to 'match self.read_queue_system' with the format of:
             case '<queue name>'
                 self.read_option_start = '<Format of how to read queue options>'
-                options = self.option_pass()
+                options, self.works = self.read_script()
                 self.obj_options.<OBJ Option function name for this queue>_options(options)
                 self.outputvariable_convert(job_directory)
         
@@ -131,7 +107,7 @@ class Script:
             self.read_option_start = ''
             read_job_name = '${}'
             read_job_id = '${}'
-            options = self.option_pass()
+            options, self.works = self.read_script()
             self.obj_options.<queue_name>_options(options)
             self.outputvariable_convert(job_name, job_id, read_job_name, read_job_id)
         =====
@@ -147,7 +123,6 @@ class Script:
                 self.submission = 'qsub'
             case None:
                 if queue_options is None:
-
                     print("No queue system was selected, therefore only queue_options['workdir'] is needed.")
 
             case _:
@@ -156,12 +131,12 @@ class Script:
         match self.read_queue_system:
             case 'slurm':
                 self.read_option_start = '#SBATCH '
-                options = self.option_pass()
+                options, self.works = self.read_script()
                 self.obj_options.slurm_options(options)
                 self.outputvariable_convert()
             case 'torque':
                 self.read_option_start = '#PBS '
-                options = self.option_pass()
+                options, self.works = self.read_script()
                 self.obj_options.torque_options(options)
                 self.outputvariable_convert()
             case None:
@@ -179,43 +154,81 @@ class Script:
         # ==========================
         self.obj_options.remove_empty_options()
 
-    def add_likwid(self, likwid_req):
-        self.likwid = True
-        self.likwid_file = impresources.files(data) / "likwid_commands.txt"
-        self.likwid_initEndSplit = -1
-        self.likwid_req = likwid_req
+    def initialise_profiling(self, profiler, profilefile):
+        """
+        initialse_profiling is used to call the define_initialise() function of the different profilers
+        Parameters
 
-    def add_prometheus(self, prometheus_req: list, prometheus_ip: str = None):
-        contains_ps = False
-        contains_pp = False
-        if prometheus_ip is not None:
-            self.prometheus_ip = prometheus_ip
-        if prometheus_req is not None:
-            for req in prometheus_req:
-                if 'PROMETHEUS_SOFTWARE' in req:
-                    contains_ps = True
-                    continue
-                if 'PROMETHEUS_PYTHON' in req:
-                    contains_pp = True
-                    continue
-                else:
-                    prometheus_req += ['export PROMETHEUS_PYTHON={}'.format(sys.executable)]
-                    contains_pp = True
-                    continue
+        ----------
+        profiler: str
+            name of the profiler to be used, must match the .py file name located in PyProfQueue.profilers
+                Currently supports: ["likwid", "prometheus"]
+        profilefile: io.TextIOWrapper
+            open profile file with write permissions.
+
+        Returns None
+        -------
+
+        """
+        if self.profiling[profiler] is None:
+            exit(f"Each profiling type, has to have a value that is not None. Profiling {profiler} "
+                 f"had a value of None")
         else:
-            prometheus_req = []
-        if not contains_pp:
-            prometheus_req += ['export PROMETHEUS_PYTHON={}'.format(sys.executable)]
-        if not contains_ps and self.prometheus_ip is None:
-            exit('When requesting prometheus profiling, prometheus_req must include "export PROMETHEUS_SOFTWARE=".')
-        self.prometheus = True
-        self.prometheus_file = impresources.files(data) / "prometheus_commands.txt"
-        self.prometheus_initEndSplit = -1
-        self.prometheus_req = prometheus_req
+            module = ".profilers."+profiler
+            current_prof = importlib.import_module(module, package="PyProfQueue")
+            current_prof.define_initialise(profilefile=profilefile,
+                                           profilerdict=self.profiling[profiler])
 
-    def option_pass(self):
-        options, self.works = self.read_script()
-        return options
+    def run_work_profiling(self, profiler, profilefile, bash_options):
+        """
+        run_work_profiling is used to call the define_run() function of the different profilers. This function makes
+        it so that the user specified bash script is executed using the chosen profiler. This can only occur once,
+        listing multiple profilers that rely on this functions will cause an error.
+
+        Parameters
+        ----------
+        profiler: str
+            name of the profiler to be used, must match the .py file name located in PyProfQueue.profilers
+            Currently supports: ["likwid", "prometheus"]
+        profilefile: io.TextIOWrapper
+            open profile file with write permissions.
+        bash_options: list
+            List of bash options to pass to the user defined bash script.
+
+        Returns None
+        -------
+
+        """
+        module = ".profilers."+profiler
+        current_prof = importlib.import_module(module, package="PyProfQueue")
+        if hasattr(current_prof, "define_run"):
+            if not self.at_execute:
+                current_prof.define_run(profilefile=profilefile,
+                                        bash_options=bash_options,
+                                        tmp_work_script=self.tmp_work_script)
+                self.at_execute = True
+            else:
+                exit(f"Multiple at execution profilers were defined. Failed when adding profiler {profiler}")
+
+    def end_profiling(self, profiler, profilefile):
+        """
+        end_profiling is used to call the define_end() function of the different profilers.
+
+        Parameters
+        ----------
+        profiler: str
+            name of the profiler to be used, must match the .py file name located in PyProfQueue.profilers
+                Currently supports: ["likwid", "prometheus"]
+        profilefile: io.TextIOWrapper
+            pen profile file with write permissions.
+
+        Returns
+        -------
+
+        """
+        module = ".profilers."+profiler
+        current_prof = importlib.import_module(module, package="PyProfQueue")
+        current_prof.define_end(profilefile=profilefile)
 
     # Queue System specifics {2}
     '''
@@ -300,6 +313,13 @@ class Script:
     '''
     # ==========================
     def outputvariable_convert(self):
+        """
+        outputvariable_convert converts the variables in a script that are queue specific for output files.
+
+        Returns None
+        -------
+
+        """
         match self.read_queue_system:
             case 'slurm':
                 match self.queue_system:
@@ -368,8 +388,20 @@ class Script:
                              'system to {}'.format(self.queue_system))
             case _:
                 exit('read_queue_system was unknown with value: {}'.format(self.read_queue_system))
+        return
     # ==========================
+
     class Options:
+        """
+        Class to read existing bash scripts, pull out the options and create an object that contains all the
+        queue options, bash options, and desired profiling methods to be injected
+
+        Parameters to initiate
+        ----------
+        queue_options: dict
+            dictionary containing all the queue options to be used. These will overwrite queue options inside the
+            user defined bash script.
+        """
         def __init__(self, queue_options: dict = None):
             self.user = None
             self.nodes = None
@@ -384,7 +416,21 @@ class Script:
             if queue_options is not None:
                 self.pass_options(queue_options)
 
-        def pass_options(self, stated_options):
+        def pass_options(self, stated_options: dict):
+            """
+            pass_options takes a dictionary containing all the queue options to be used. These will overwrite queue
+            options in the user defined bash script.
+
+            Parameters
+            ----------
+            stated_options: dict
+                dictionary containing all the queue options to be used. These will overwrite queue options inside the
+                user defined bash script.
+
+            Returns None
+            -------
+
+            """
             for key, value in stated_options.items():
                 match key:
                     case 'user':
@@ -409,13 +455,21 @@ class Script:
                         self.output = value
                     case _:
                         exit('Unknown option value in key: {}, with value: {} passed to pass_options'.format(key, value))
+            None
 
         def remove_empty_options(self):
+            """
+            remove_empty_options removes all empty options from the Options object.
+
+            Returns None
+            -------
+            """
             empty_attributes = [a for a in dir(self) if (not a.startswith('__') and
                                                          not callable(getattr(self, a)) and
                                                          getattr(self, a) is None)]
             for attribute in empty_attributes:
                 delattr(self, attribute)
+            return
 
         # Queue System specifics {3}
         '''
@@ -609,7 +663,21 @@ class Script:
         # ==========================
 
     def change_options(self, queue_options: dict):
+        """
+        change_options allows users to change the options they specified, after initialising their object.
+
+        Parameters
+        ----------
+        queue_options: dict
+            dictionary containing all the queue options to be used. These will overwrite queue options inside the
+            user defined bash script.
+
+        Returns None
+        -------
+
+        """
         self.obj_options.pass_options(queue_options)
+        return
 
     # Queue System specifics {4}
     '''
@@ -716,6 +784,13 @@ class Script:
     # ==========================
 
     def read_script(self):
+        """
+        read_script reads the user defined bash script and returns the options as a dictionary, and all the work
+        that is to be run as a list of strings.
+
+        Returns options (dict), work (list[str])
+        -------
+        """
         options = {}
         work = []
         with open(self.work_script, 'r') as script:
@@ -740,22 +815,67 @@ class Script:
         return options, work
 
     def create_workfile(self):
+        """
+        create_workfile creates the work file based on the user defined bash script by simply removing the options.
+        This is not used, when the work script has no queue options.
+        Returns
+        -------
+
+        """
         with open(self.tmp_work_script, 'w') as workfile:
             workfile.seek(0)
             for line in self.works:
                 workfile.write(line)
         return
 
+
+    def add_options(self, profilefile):
+        """
+        add_options adds the queue options from the Script object to the profile file.
+        Parameters
+        ----------
+        profilefile: io.TextIOWrapper
+            open profile file with write access.
+        Returns None
+        -------
+
+        """
+        to_write = [a for a in dir(self.obj_options) if not a.startswith('__') and
+                    not callable(getattr(self.obj_options, a))]
+        for option in to_write:
+            if option is not None and option != 'workdir':
+                profilefile.write(self.option_start + self.option_converter(option) +
+                                  self.obj_options.__getattribute__(option) + '\n')
+        profilefile.write('\n')
+        return
+
     def create_profilefile(self, tmp_work_script='./tmp_workfile.sh',
                            tmp_profile_script='./tmp_profilefile.sh',
                            bash_options=['']):
+        """
+        create_profilefile uses the attributes of the Script object, and creates the temporary profile file that will
+        be submitted to the queue on behalf of the user.
+
+        Parameters
+        ----------
+        tmp_work_script: str
+            path and name of the temporary work file that should be created. This file is the user defined bash script
+            with any queue options removed.
+        tmp_profile_script: str
+            path and name of the temporary profile file that should be created. This will contain any profiling
+            initialisations, calls and terminations needed to profile the user defined work with the desired profiling
+            software.
+        bash_options: list[str]
+            list of bash options that should be passed to the user defined bash script.
+        Returns None
+        -------
+        """
         if self.read_queue_system is not None:
             self.tmp_work_script = tmp_work_script
             self.create_workfile()
         else:
             self.tmp_work_script = self.work_script
         self.tmp_profile_script = tmp_profile_script
-
         with open(self.tmp_profile_script, 'w') as profilefile:
             profilefile.seek(0)
             profilefile.write('#!/bin/bash\n')
@@ -768,129 +888,24 @@ class Script:
             profilefile.write('fi\n')
             profilefile.write('cd ${WORKING_DIR}\n')
             profilefile.write('\n')
+            if self.profiling is not None:
+                for key in self.profiling.keys():
+                    self.initialise_profiling(key, profilefile)
 
-            if self.likwid:
-                self.init_likwid(profilefile)
+                for key in self.profiling.keys():
+                    self.run_work_profiling(key, profilefile, bash_options)
 
-            if self.prometheus:
-                self.init_prometheus(profilefile)
+                if not self.at_execute:
+                    self.run_work(profilefile, bash_options)
 
-            if self.likwid:
-                self.run_likwid(profilefile, bash_options)
+                for key in self.profiling.keys():
+                    self.end_profiling(key, profilefile)
             else:
                 self.run_work(profilefile, bash_options)
-
-            if self.prometheus:
-                self.end_prometheus(profilefile)
-
-            if self.likwid:
-                self.end_likwid(profilefile)
-
         return
 
-    def add_options(self, profilefile):
-        to_write = [a for a in dir(self.obj_options) if not a.startswith('__') and
-                    not callable(getattr(self.obj_options, a))]
-        for option in to_write:
-            if option is not None and option != 'workdir':
-                profilefile.write(self.option_start + self.option_converter(option) +
-                                  self.obj_options.__getattribute__(option) + '\n')
-        profilefile.write('\n')
-
     def run_work(self, profilefile, bash_options=['']):
         profilefile.write('bash {} {}\n'.format(self.tmp_work_script,
                                                 ' '.join(str(x) for x in bash_options)))
         profilefile.write('\n')
         return
-
-    def init_likwid(self, profilefile):
-        profilefile.write('# Likwid initialisation declarations\n')
-        for i in self.likwid_req:
-            profilefile.write(i)
-            profilefile.write('\n')
-        profilefile.write('\n')
-        profilefile.write('export LIKWID_RUNNING_DIR=${WORKING_DIR}/Likwid\n')
-        with open(self.likwid_file, 'r') as likwid_file:
-            for number, line in enumerate(likwid_file):
-                if line == '# *=*\n':
-                    self.likwid_initEndSplit = number + 1
-                    break
-                profilefile.write(line)
-        profilefile.write('# Likwid initialisation done\n')
-        profilefile.write('\n')
-
-    def run_likwid(self, profilefile, bash_options=['']):
-        profilefile.write('likwid-perfctr -g MEM_DP -f bash {} {}\n'.format(self.tmp_work_script,
-                                                                            ' '.join(str(x) for x in bash_options)))
-        profilefile.write('\n')
-        return
-
-    def end_likwid(self, profilefile):
-        profilefile.write('# Likwid final steps declarations\n')
-        with open(self.likwid_file, 'r') as likwid_file:
-            for line in itertools.islice(likwid_file, self.likwid_initEndSplit, None):
-                profilefile.write(line)
-        profilefile.write('# Likwid final steps done\n')
-        profilefile.write('\n')
-        return
-
-    def init_prometheus(self, profilefile):
-        profilefile.write('# Prometheus initialisation declarations\n')
-        if self.prometheus_ip is None:
-            profilefile.write("export PROMETHEUS_IP=http://localhost:9090\n")
-        else:
-            profilefile.write("export PROMETHEUS_IP={}\n".format(self.prometheus_ip))
-        for i in self.prometheus_req:
-            profilefile.write(i)
-            profilefile.write('\n')
-        profilefile.write('export PROMETHEUS_RUNNING_DIR=${WORKING_DIR}/Prometheus\n')
-        scrape_path = str(impresources.path(data, 'read_prometheus.py'))[:-19]
-        profilefile.write('export PROFILE_SCRAPE={}\n'.format(scrape_path))
-        profilefile.write('\n')
-        final_init_indicator = 2
-        if self.prometheus_ip is None:
-            read_indicators = [0, 1, 2]
-        else:
-            read_indicators = [0, 2]
-        indicator = 0
-        with open(self.prometheus_file, 'r') as prometheus_file:
-            for number, line in enumerate(prometheus_file):
-                if line == '# *=*\n' and indicator >= final_init_indicator:
-                    self.prometheus_initEndSplit = number + 1
-                    break
-                elif line == '# *=*\n':
-                    indicator += 1
-                    continue
-                if indicator in read_indicators:
-                    profilefile.write(line)
-        print(indicator)
-        print(self.prometheus_initEndSplit)
-        profilefile.write('# Prometheus initialisation done\n')
-        profilefile.write('\n')
-
-    def end_prometheus(self, profilefile):
-        profilefile.write('# Prometheus final steps declarations\n')
-        if self.prometheus_ip is None:
-            read_indicators = [0, 1, 2]
-        else:
-            read_indicators = [0, 2]
-        indicator = 0
-        with open(self.prometheus_file, 'r') as prometheus_file:
-            for line in itertools.islice(prometheus_file, self.prometheus_initEndSplit, None):
-                if line == '# *=*\n':
-                    indicator += 1
-                    continue
-                if indicator in read_indicators:
-                    profilefile.write(line)
-        profilefile.write('# Prometheus final steps done\n')
-        profilefile.write('\n')
-
-    def existing_init_prometheus(self, profilefile):
-        profilefile.write('# Prometheus initialisation for scraping existing instance')
-        for i in self.prometheus_req:
-            profilefile.write(i)
-            profilefile.write('\n')
-        profilefile.write('export PROMETHEUS_RUNNING_DIR=${WORKING_DIR}/Prometheus\n')
-
-    def existing_end_prometheus(self, profilefile):
-        profilefile.write('# Prometheus final steps for scraping existing instance')
